use anyhow::{
    Context,
    Result,
};
use clap::{
    Parser,
};
use feed_rs::{
    parser::parse,
    model::Feed,
};

use reqwest::{
    IntoUrl,
    Url,
};

#[derive(Parser)]
#[command(version, about, long_about = None)]
struct Cli {
    /// Path to file containing feeds
    #[arg(long, default_value_t = String::from("feeds.txt"))]
    feeds: String,
    /// Path to output file
    #[arg(long, default_value_t = String::from("index.html"))]
    out: String,
}

struct FeedItem {
    link: String,
    title: String,
    time: Option<chrono::DateTime<chrono::Utc>>,
}

fn fetch_feed(url: impl IntoUrl) -> Result<Feed> {
    let content = reqwest::blocking::get(url)?.bytes()?;
    Ok(parse(&content[..])?)
}

struct FeedFile {
    name: String,
    urls: Vec<Url>,
}

fn read_feed_file(path: &str) -> Result<FeedFile> {
    let mut name = String::from(path);
    let mut urls = Vec::new();

    for line in std::fs::read_to_string(path)?.lines() {
        let line = line.trim().to_string();
        if line == "" {
            continue;
        }
        if line.starts_with("# name: ") {
            name = String::from(&line[8..]);
            continue;
        }
        if line.starts_with("#") {
            continue;
        }
        let url = Url::parse(&line)?;
        println!("{}", url);

        urls.push(url);
    }

    Ok(FeedFile{
        name: name,
        urls: urls,
    })
}

fn main() -> Result<()> {
    let cli = Cli::parse();

    let mut feed_items = Vec::new();

    let feed_file = read_feed_file(&cli.feeds)?;

    for feed_url in feed_file.urls {
        feed_items.extend(fetch_feed(feed_url)?.entries);
    }

    let mut items: Vec<FeedItem> = feed_items.into_iter().map(|item| FeedItem {
        link: item.links.first().unwrap().href.clone(),
        title: item.title.unwrap().content.clone(),
        time: match item.published {
            Some(time) => Some(time),
            None => item.updated,
        },
    }).collect();
    items.sort_by(|a, b| a.time.unwrap_or(chrono::DateTime::UNIX_EPOCH).cmp(&b.time.unwrap_or(chrono::DateTime::UNIX_EPOCH)).reverse());

    let mut out = String::new();

    out.push_str("<!DOCTYPE html>\n");
    out.push_str("<html>\n");
    out.push_str("<head>\n");
    out.push_str(&format!("<title>{} - rainbowrss</title>", feed_file.name));
    out.push_str("</head>\n");
    out.push_str("<body>\n");
    out.push_str(&format!("<h1>{}</h1>", feed_file.name));

    out.push_str("<ul>\n");

    for item in items {
        out.push_str(&format!("<li><a href=\"{}\">[l]</a> {} {}</li>\n", item.link, item.title, match item.time{
            Some(time) => format!("({})", time),
            None => "".to_string(),
        }));
    }

    out.push_str("</ul>\n");

    out.push_str(&format!("<small>Last updated: {}</small>", chrono::Utc::now()));

    out.push_str("</body>\n");
    out.push_str("</html>\n");

    std::fs::write(&cli.out, out)
        .context("Failed to write output to file")?;

    Ok(())
}
